package current_rpc

  1. Overview
  2. Docs

Module Current_rpcSource

Sourcemodule Job : sig ... end

Client-side API to contact a job service.

Sourcemodule Engine : sig ... end

Client-side API to contact an engine service.

Sourcemodule S : sig ... end

Module signatures for the RPC system.

Sourcemodule Impl (Current : S.CURRENT) : sig ... end

This is used on the server-side to provide access to the OCurrent Engine. Create an instance of the functor with module Rpc = Current_rpc.Impl(Current). We use a functor here just to avoid having Current_rpc depend on Current, which would be annoying for RPC clients.

Sourcemodule Client : module type of Client

Unified RPC client with cmdliner integration. See Client for details on embedding in applications.

Sourcemodule Impl_with_db (Current : S.CURRENT) (Db : S.DB) : sig ... end

Functor that takes an explicit Db module for job history queries. This is primarily useful for testing with mock databases. For normal use, prefer Impl which uses Current_cache.Db.