package current_rpc

  1. Overview
  2. Docs

Module Current_rpc.Impl_with_dbSource

Functor that takes an explicit Db module for job history queries. This is primarily useful for testing with mock databases. For normal use, prefer Impl which uses Current_cache.Db.

Parameters

module Db : S.DB

Signature

Sourceval job : engine:Current.Engine.t -> Job.id -> Job.t

job ~engine id is a Cap'n Proto job service backed by engine.

engine e is a Cap'n Proto engine service backed by e.