Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Current_rpcSourceThis is used on the server-side to provide access to the OCurrent Engine. Create an instance of the functor with module Rpc = Current_rpc.Impl(Current). We use a functor here just to avoid having Current_rpc depend on Current, which would be annoying for RPC clients.
Unified RPC client with cmdliner integration. See Client for details on embedding in applications.